#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fcntl.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/time.h>

#include <linux/timer.h>

#include <linux/ptrace.h>			/* used to get the pt_regs structure definition */
#include <linux/cdev.h>
#include <linux/slab.h>				/* for kmalloc() and kfree() */
#include <linux/vmalloc.h>    // for tracing phase
#include <linux/mm.h>				  /* for GFP_KERNEL macro */

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/system.h>
//Include for Trace
#include <linux/param.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include "rgrs_lib_trace.h"
#include "pmc_asm.h"

//Include for DVFS
#include <linux/cpufreq.h>
#include <asm/semaphore.h>
#define TRY_NUM 10		// maximum # of trial for DVFS


#define KERNELTIMER_WRITE_ADDR 0x0378
#define TIME_STEP (3*HZ/10)

#define CMD_START 0xcc
#define CMD_STOP  0xbb

#define TRACE_MAJOR 230
//#define TRACE_MINOR 0
#define DEV_NAME "pmutrace"

#define DEBUG

#ifdef DEBUG
#define pr_debug(fmt,arg...)	printk(KERN_DEBUG fmt,##arg)
#else
#define pr_debug(fmt,arg...)	do { } while(0)
#endif


static void pmc_stop(void);
static void pmc_setup(int);

unsigned long long int PMU1, PMU2, INST;
char sndbuf[40];
char *buff;
unsigned int th_cntp = 0, th_cntm = 0;

unsigned int monitor_flag = 0,  trigger_flag = 0;;

struct trace_result_type *trace_result;

struct trace_result_type count_read;

typedef struct{
	struct timer_list timer;
} __attribute__ ((packed)) KERNEL_TIMER_MANAGER;

static KERNEL_TIMER_MANAGER *ptrmng = NULL;

void kerneltimer_timeover(unsigned long arg);

static void pmi_handler(void)
{
	int cpu, pmi_num, set_freq;
	pmc_uint64_t buf64;	
	//pr_debug("pmi_handler called\n");
  // get current operating CPU #
	cpu = smp_processor_id();
	pmi_num = trace_result[cpu].n_pmi;
	
	// stop counting
	pmc_stop();

	// read pmc counter readings  
	PMC_ASM_READ_PMC(1, buf64.bits64);
	printk("buf PMU1:%llu:", buf64.bits64);
	PMU1 = buf64.bits64;
	//trace_result[cpu].PMU1[pmi_num] = buf64.bits64;
	
  PMC_ASM_READ_PMC(0, buf64.bits64);
  printk("buf PMU2:%llu:", buf64.bits64);
  //trace_result[cpu].PMU2[pmi_num] = buf64.bits64;
  PMU2 = buf64.bits64;
  
  PMC_ASM_READ_MSR(0, buf64.bits64);
  printk("buf INST: :%llu:\n", buf64.bits64);
  INST = buf64.bits64;
  
  if (INST > threshold)
  	th_cntp++;
  	
  if (th_cntp > 2)
  	{
 			set_freq = cpufreq_get(cpu);				// get frequency
  		th_cnt = 0;

  		if(set_freq!= 2158000 && set_freq != 1328000)
  		change_freq(2158000);
  		else if(set_freq!=2158000 && set_freq != 1660000)
  		change_freq(1660000);
  	}
  else if(th_cntm > 2)
  	{
  		set_freq = cpufreq_get(cpu);				// get frequency
  		th_cntm = 0;
  		change_freq(1328000);
  	}
  	
  //trace_result[cpu].INST[pmi_num] = buf64.bits64;
  		
	pmc_setup(1);
}

static void pmc_setup(int flag)  
{	  

	PMC_ASM_WRITE_MSR(PMC_COUNTER0_MSR, 0);  
	// memory bus #, counter0 (0xc2, 1) = 0 
	PMC_ASM_WRITE_MSR(PMC_COUNTER1_MSR, 0);  
	
	//PMU2 in printk
 	PMC_ASM_WRITE_MSR(PMC_EVNTSEL0_MSR, UOPS_RETIRED_EVENT);	
 	PMC_ASM_WRITE_MSR(PMC_EVNTSEL1_MSR, BUS_TRAN_MEM_EVENT); 	  
  // clear
 	PMC_ASM_WRITE_MSR(FIXED_CTR0, 0);  
 	// Fixed count enable	
 	PMC_ASM_WRITE_MSR(FIXED_CTR_CTRL, FIXED_INST);  
}

static void pmc_stop(void)
{
	//pr_debug(KERN_WARNING "CPU %d, pmc_stop()\n", smp_processor_id());
  // Number of instruction retired, p.A-62
  // PerfEvtSel0 (EVNTSEL0, 0x186) = 0
 	PMC_ASM_WRITE_MSR(PMC_EVNTSEL0_MSR, 0);

 	// Number of completed memory transactions, p. A-57
 	// PerfEvtSel1 (EVNTSEL1, 0x187) = 0
 	PMC_ASM_WRITE_MSR(PMC_EVNTSEL1_MSR, 0); 	 		
 	// Fixed count disable
 	PMC_ASM_WRITE_MSR(FIXED_CTR_CTRL, 0);   
}
void initialize_trace(void)
{
		//PMU register initialize 
		pr_debug("initialize_trace\n");		
}

void setup_pmu_events(void)
{
		//set specific PMU events
		pr_debug("setup_pmu_events\n");
		pmc_setup(1);
}

void core2duo_pmu_stop(void)
{
		int i, cpu; 
		cpu = smp_processor_id();
		//Stop PMU counting
		monitor_flag = 0;
		trigger_flag = 2;
		pr_debug("core2duo_pmu_stop, monitor_flag = %d\n",monitor_flag);
		/*
		for(i=0;i<trace_result[cpu].n_pmi; i++)
		{
			 printk(KERN_INFO ">>>> CPU:%u, pmi_no:%u, INST :%llu, PMU1:%llu, PMU2:%llu\n", 
         cpu, i, trace_result[cpu].INST[i],
         trace_result[cpu].PMU1[i], trace_result[cpu].PMU2[i]);
    }*/
		//copy_to_user(sendbuf,"exit",5);
}

void change_freq(int target_freq)
{
	struct cpufreq_policy *policy;
	int set_freq, try_num, cpu;
	
	try_num = 0;
	
	printk(KERN_INFO "SetDvfs: CPU%d change_freq(), call cpufreq_set(%d)\n", 
                                   smp_processor_id(), target_freq);
  cpu = smp_processor_id();
again:
	++try_num;
	
	policy = cpufreq_cpu_get(cpu);  // get cpu# policy      
	policy->cpu = cpu;
	
	// /usr/src/linux/drivers/cpufreq/cpufreq.c, line 1410
	cpufreq_driver_target(policy,target_freq,0);
	// /usr/src/linux/drivers/cpufreq/cpufreq.c, line 1126	
	set_freq = cpufreq_get(cpu);				// get frequency

  printk(KERN_INFO "SetDvfs: CPU%d try %d, target_freq:%d, set_freq: %d\n", 
			                    cpu, try_num, target_freq, set_freq);
  if ((target_freq != set_freq) && try_num < TRY_NUM) {
  	  cpu = cpu ^ 1;
  		goto again;
  }
}

void kerneltimer_registertimer(KERNEL_TIMER_MANAGER *pdata, unsigned long timeover)
{
	init_timer(&(pdata->timer));
	pdata->timer.expires  = get_jiffies_64() + timeover;
	pdata->timer.data     = (unsigned long) pdata;
	pdata->timer.function = kerneltimer_timeover;
	add_timer(&(pdata->timer));
	//pr_debug("kerneltimer_registertimer called!!\n");
}

int core2duo_pmu_start(void)
{
		monitor_flag = 1;
		//pr_debug("core2duo_pmu_start, monitor_flag = %d\n",monitor_flag);
		return 0;
}

void kerneltimer_timeover(unsigned long arg)
{
	KERNEL_TIMER_MANAGER *pdata = NULL;
	//pr_debug("arg is %ld\n",arg);
	
	if(arg)
		{
			pdata = (KERNEL_TIMER_MANAGER *) arg;
			//pr_debug("Led inverted!!,monitor_flag = %d\n",monitor_flag);
			kerneltimer_registertimer(pdata, TIME_STEP);
		}
		if(monitor_flag == 1 )
			{
			//pr_debug("TRACING PMU--\n");
			pmi_handler();
		
			trigger_flag = 1;
			}
		
}

int trace_ioctl(struct inode *inode, struct file *filp, unsigned int ioc_cmd, unsigned long arg)
{
	pr_debug(">trace_ioctl(), cmd:%2x\n",ioc_cmd);
	
	switch(ioc_cmd) 
	{
			case CMD_START:
				//initialize_trace();
				pr_debug("CMD_START, initialize_trace finished\n");
				//setup_pmu_events(); //move to module init part
				pr_debug("CMD_START, setup_pmu_events finished\n");
				core2duo_pmu_start();
				pr_debug("CMD_START, core2duo_pmu_start finished\n");
				return(0);
			case CMD_STOP:
				core2duo_pmu_stop();
				return(0);
			default:
				return(-ENOTTY);
	}
}

static int trace_device_open(struct inode * inode, struct file * file)
{
	return(0);
}

static int trace_device_close(struct inode * inode, struct file * file)
{
	
	/* shutdown */
	//trace_shutdown();
	return (0);
}

static ssize_t trace_device_read(struct file * file, char * buf, size_t cmd, loff_t * unused)
{
	char sndbuf[70];
	sprintf(sndbuf,"client_event1_%llu_event2_%llu",INST,PMU1);
	
	PMU1 = 0;
	PMU2 = 0;
	INST = 0;
		
	printk("%s\n",sndbuf);	
	pr_debug("[FUNC] trace_device_read called\n");
	copy_to_user(buf,sndbuf,sizeof(sndbuf));
	return 0;
}
		

struct file_operations trace_device_fops = {
	.owner = THIS_MODULE,
	.read  = trace_device_read,
	.ioctl = trace_ioctl,
	.open  = trace_device_open,
	.release = trace_device_close,
};

//struct cdev trace_cdev;

int trace_device_init(void)
{
		int result;
		pr_debug("Initialization trace_device_init\n");
  	//Start signal for WT-210
  	outb(0x01, PARALLEL_WRITE_PORT);
  	mdelay(5);
  	outb(0x00, PARALLEL_WRITE_PORT);
  	mdelay(5);
  	outb(0x01, PARALLEL_WRITE_PORT);
  	
		//Setup timer
		ptrmng = kmalloc(sizeof(KERNEL_TIMER_MANAGER), GFP_KERNEL);
		if(ptrmng == NULL) return -ENOMEM;
		memset(ptrmng, 0, sizeof(KERNEL_TIMER_MANAGER));
		kerneltimer_registertimer(ptrmng, TIME_STEP);
		
		//Register character device	
		result = register_chrdev(TRACE_MAJOR,DEV_NAME, &trace_device_fops);
		if(result < 0) 
  	{
			pr_debug("pmutrace : can't get major %d\n", TRACE_MAJOR);
			return result;
		}
		
		trace_result = vmalloc(N_CPU*sizeof(struct trace_result_type));
		setup_pmu_events();
		return 0;
}
	
void trace_device_exit(void)
{
	pr_debug("Called trace_device_exit\n");
	unregister_chrdev(TRACE_MAJOR,DEV_NAME);
	
	if(ptrmng !=NULL)
		{
			del_timer(&(ptrmng->timer));
			kfree(ptrmng);
		}
		//outb(0x00, KERNELTIMER_WRITE_ADDR);
		vfree(trace_result);
}	

module_init(trace_device_init);
module_exit(trace_device_exit);
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Hee-Sung Yun <iloveyr2@gmail.com>");
MODULE_DESCRIPTION("PMU trace driver for E7520 development board(Core2Duo T7200)");
